# Copyright (C) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from copy import deepcopy
from generator import Problem, ProblemSet

amax_def_args = {'--type'  : 'H',
                 '--dtype' : 'S',
                 '--init'  : 'hpl',
                 }

lengths = {

    'amax_example': [
        (64, 64),
        (128, 128),
    ],

    'amax_set_1': [
        (16,    1024  ),
        (16,    8192  ),
        (16,    65536 ),
        (2048,  1024  ),
        (2048,  8192  ),
        (2048,  65536 ),
        (8192,  1024  ),
        (8192,  8192  ),
        (8192,  65536 ),
        (16,    16384 ),
        (32,    16384 ),
        (16,    2048  ),
        (32,    2048  ),
        (16,    4096  ),
        (32,    4096  ),
    ],
}

# Suite definitions
def api_overhead():
    """API overhead"""

    # iteraion 10000 takes too long time for "getAll", 1000 is enough
    problemlist = [Problem(args={"--cold_iters" : "10" , "--iters" : "1000"})]
    yield ProblemSet(benchType="api_overhead", name="benchset_1", problems=problemlist)

def amax_example():
    """AMAX example."""

    problemlist = []

    for length in lengths['amax_example']:
        args = deepcopy(amax_def_args)
        args.update({'--m': str(length[0]),
                     '--n': str(length[1])})
        problemlist.append(Problem(args=args))

    yield ProblemSet(benchType="amax", name="example", problems=problemlist)

def amax_set_1():
    """AMAX benchset 1."""

    problemlist = []

    for length in lengths['amax_set_1']:
        args = deepcopy(amax_def_args)
        args.update({'--m': str(length[0]),
                     '--n': str(length[1])})
        problemlist.append(Problem(args=args))

    yield ProblemSet(benchType="amax", name="benchset_1", problems=problemlist)

def matmul_set_1():
    """gemm benchset 1"""

    problemlist = [Problem(args={"--log_function_name" : "" , "--yaml" : "matmul_probset1_bench.yaml"})]
    yield ProblemSet(benchType="matmul", name="benchset_1", problems=problemlist)

def matmul_set_2():
    """gemm benchset 2"""

    problemlist = [Problem(args={"--log_function_name" : "" , "--yaml" : "matmul_probset2_bench.yaml"})]
    yield ProblemSet(benchType="matmul", name="benchset_2", problems=problemlist)

def all():
    """all routine benchmarks"""

    yield from api_overhead()
    yield from amax_set_1()
    yield from matmul_set_1()
    yield from matmul_set_2()



# Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging

from dataclasses import dataclass, field
from pathlib import Path as path
from typing import Dict, List, Mapping, Generator

top = path(__file__).resolve().parent

@dataclass
class Problem:
    benchType: str = None
    args: Dict[str, str] = field(default_factory=dict)  # storing all arguments

@dataclass
class ProblemSet:
    benchType: str
    name: str
    problems: List[Problem]

    def __iter__(self):
        idx = 0
        while idx < len(self.problems):
            yield self.problems[idx]
            idx += 1

    def get_problemset_name(self):
        return self.benchType + "_" + self.name

    def generate_problems(self):
        for p in self.problems:
            p.benchType = self.benchType
            yield p

def load_suite(suite):
    """Load performance suite from suites.py."""

    tdef = top / 'suites.py'
    logging.info(f'loading suites from {tdef}')
    code = compile(tdef.read_text(), str(tdef), 'exec')
    ns = {}
    exec(code, ns)
    return ns[suite]

@dataclass
class SuiteProblemGenerator:
    suite_names: List[str]
    suites: Mapping[str, Generator[ProblemSet, None,
                                   None]] = field(default_factory=dict)

    def __post_init__(self):
        for name in self.suite_names:
            self.suites[name] = load_suite(name)

    def generate_problemSet(self):
        for g in self.suites.values():
            yield from g()

@dataclass
class ShellScriptProblemGenerator:
    sh_filename: str = None
    bench_exec: str = None
    benchCMDs: List[str] = field(default_factory=list)

    def __post_init__(self):
        # print("sh file: " + str(self.sh_filename))
        # print("bench_exec: " + str(self.bench_exec))
        # load file and readline
        try:
            with open(self.sh_filename, 'r') as sh_file:
                # Iterate over each line in the file
                for cmd in sh_file:
                    cmd = cmd.strip()
                    # skip invalid line
                    if cmd.startswith("#") or cmd.count("hipblaslt-bench") == 0:
                        continue
                    # --verfiy will output other values, not supported yet
                    if cmd.count("verfiy") > 0 or cmd.count("-v") > 0:
                        print(f'--verify or -v is not supported, skip bench.')
                        continue
                    # TODO- not supported for offline tuning
                    if cmd.count("requested_solution") > 0:
                        print(f'--requested_solution is not supported, skip bench.')
                        continue

                    cmd = cmd.replace("./hipblaslt-bench", str(self.bench_exec))
                    # print("parsed cmd: " + cmd)
                    self.benchCMDs.append(cmd)

        except FileNotFoundError:
            print(f'Error: The file {self.sh_filename} was not found.')

        except Exception as e:
            print(f'An error occurred: {e}')

    def iterate_cmd(self):
        for cmd in self.benchCMDs:
            yield cmd



# Copyright © Advanced Micro Devices, Inc., or its affiliates.
# SPDX-License-Identifier: MIT

"""
This module is used to create git info file,
if it is not found in output_folder_location specified by the end user.
"""


from pathlib import Path
import git
import os


def get_merge_id(git_log: list):

    """
    This method is used to get merge id and pull id from a git log.

    Args
    ------
    git_log : string consisting of all the information about commit id.

    Returns
    -------
    merge_id, pull_id : Merge id , Pull id for the latest commit.

    """

    merge_id = ""
    pull_id = ""

    try:
        merge_id = git_log[0][7:]
        for i in git_log:
            if "merge pull request #" in i.lower():
                pull_id = i.lower().split()[3][1:]

    except Exception as e:
        print("No Merge requests found.")

    return merge_id, pull_id


def create_github_file(filename: str, dir_name: str = None) -> str:
    """
    This method creates a github information file in the specified directory.

    This function creates four different parameters of a
    git repo { Branch Name , Merge Id ,  Pull Id ,  Git Hash }

    To run this function you must be in a project/package or in any git repo folder
    (or) you need to give the location of a git repo when prompted.

    Args
    -----
    filename : Absolute filename in which the path is to be created
    dir_name : Absolute path where the git hash to be generated (optional)

    Returns
    -------
    filename : Absolute filename of created file

    Raises
    ------
    InvalidGitRepositoryError : If the package is running outside of a git repo
    IOError (Is a directory error) : If the specified filename ( input ) should be a file
                                     and not a directory

    Usage
    -----
    CLI syntax
    ----------
    linux : python3 -m pts_amd create_git_info < filename >
    windows : python -m pts_amd create_git_info < filename >

    """

    # Declaring the git info parameters as empty strings
    branch_name = "None"
    merge_id = "None"
    pull_id = "None"
    git_hash = "None"

    try:
        cwd = os.getcwd()
        if dir_name:
            if os.path.isdir(Path(dir_name)):
                os.chdir(Path(dir_name))
        try:
            repo = git.Repo(Path(dir_name))
            os.chdir(cwd)
        except (git.InvalidGitRepositoryError, git.exc.NoSuchPathError, Exception) as err:
            print(err)
            os.chdir(cwd)
            repo = git.Repo(search_parent_directories=True)
        # Retrieving the values, when package/project is running inside any git repo
        # Check if the repo is in a detached HEAD state
        if repo.head.is_detached:
            print("Repository is in a detached HEAD state.")
            # You can use repo.head.commit to get the current commit
            git_hash = repo.head.commit.hexsha
            branches_output = repo.git.branch('--contains', git_hash)
            branches = [
                branch.strip().lstrip("* ")  # Remove leading "* " for the current branch
                for branch in branches_output.split("\n")
                if branch.strip()  # Exclude empty lines
            ]
            branch_name = branches[0] if branches else "None"
        else:
            # Safe to access active_branch
            branch_name = repo.active_branch.name
            git_hash = repo.head.object.hexsha
            git_logs = repo.git.log("--grep=Merge", "--max-count=1")
            if len(git_logs) > 0:
                merge_id, pull_id = get_merge_id(git_logs.split("\n"))

    except (git.InvalidGitRepositoryError, git.exc.NoSuchPathError):
        print(
            "Could not find any git files from github repo in the specified output folder location."
        )
        return filename

    # If all parameters are captured successfully then those parameters are written in a file.
    try:
        with open(filename, "w") as file:
            file.write(f"Branch Name : {branch_name}\n")
            file.write(f"Merge Id : {merge_id}\n")
            file.write(f"Pull Id : {pull_id}\n")
            file.write(f"Git Hash : {git_hash}\n")

    except Exception as error:
        print(error)
        return filename

    return filename



# Copyright © Advanced Micro Devices, Inc., or its affiliates.
# SPDX-License-Identifier: MIT

from pathlib import Path
import re
import socket
import subprocess
import tempfile


def run(cmd):
    p = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return p.stdout.decode("ascii")


def search(pattern, string):
    m = re.search(pattern, string, re.MULTILINE)
    if m is not None:
        return m.group(1)
    return None


def _subprocess_helper(cmd: list) -> tuple:
    """
    This is a helper method which runs a command line argument
    and returns the result in text format.

    Args
    ----
    cmd : A list of commands with relevant options

    Returns
    -------
    success : A boolean to indicate whether the command was executed succcesfully
    info : The result of the command in text format

    """

    fout = tempfile.TemporaryFile(mode="w+")
    ferr = tempfile.TemporaryFile(mode="w+")
    success = False
    info = ""

    try:
        process = subprocess.Popen(cmd, stdout=fout, stderr=ferr)
        process.wait()
        fout.seek(0)
        info = fout.read()
        success = True
    except subprocess.CalledProcessError as process_error:
        info = "None"
    except FileNotFoundError as not_found_err:
        info = "None"
    finally:
        fout.close()
        ferr.close()

    return success, info


def get_cuda_info() -> str:
    """
    This function gets the cuda version if available. Else, returns "None".

    Args
    ----
    None

    Returns
    -------
    info : A string containing the CUDA version information

    """

    path = Path("/usr/local/cuda/version.txt")
    if path.exists():
        info = path.read_text()

    else:
        success, info = _subprocess_helper(["nvidia-smi"])

        if success:
            info = re.split(r"\{2,}", info.split("\n")[2])[-2]
            return info

        else:
            success, info = _subprocess_helper(["nvcc", "--version"])

            if success:
                info = "\n".join(info.split("\n")[-3:-1])
                return info

    return info


def get_gpu_info(devicenum: int = 0) -> str:
    """
    This function gets the NVIDIA GPU infomration if available. Else, returns "None".

    Args
    ----
    devicenum : A integer denoting the device number

    Returns
    -------
    info : A string containing the NVIDIA GPU information

    """

    success, info = _subprocess_helper(
        [
            "nvidia-smi",
            "--query-gpu=name",
            "--format=csv,noheader",
            "-i",
            str(devicenum),
        ]
    )

    return info


def get_device_info() -> str:

    """

    This function returns architecture_name and internal_product_name from rocminfo.
    If not found returns blank output.

    """

    success, info = _subprocess_helper("rocminfo")
    architecture_name = ""
    internal_product_name = ""

    if info != "None":

        output = info.split("\n")
        start_num = 0
        for i, row in enumerate(output):
            if "Agent 2" in row:
                start_num = i

        if start_num != 0:

            for i in range(start_num, len(output)):
                if "Name" in output[i]:
                    s = search(r"Name:\s*(.*?)$", output[i])
                    if s.startswith("gfx"):
                        architecture_name = s
                        break

            for i in range(start_num, len(output)):
                if "Marketing Name" in output[i]:
                    internal_product_name = search(
                        r"Marketing Name:\s*(.*?)$", output[i]
                    )
                    break

    return architecture_name, internal_product_name



def get_sbios_info() -> str:

    """

    This function returns architecture_name and internal_product_name from rocminfo.
    If not found returns blank output.

    """

    sbios_info = 'None'
    try:
        sbios_info = Path('/sys/class/dmi/id/bios_vendor').read_text().strip() + Path(
        '/sys/class/dmi/id/bios_version').read_text().strip()
    except Exception:
        sbios_info = 'None'
    return sbios_info


def get_machine_specs(filename: str, devicenum: int = 0):
    """
    This method is used to get the machine specifications and
    store it in the specified location as a text file.

    Args
    ----------
    filename : Absolute filepath where the file is to be created
    devicenum : Device number { optional }

    Returns
    -------
    filename : Absolute filepath of the created specs file

    Raises
    ------
    IOError (Is a directory error) : If the specified filename ( input ) should be a file
                                     and not a directory

    Usage
    -----
    CLI syntax
    ----------
    linux : python3 -m pts_amd create_specs_info < filename >
    windows : python -m pts_amd create_specs_info < filename >

    """

    try:
        # Getting the specs information of current device
        # If the specified information is not found , then the value is stored as None
        cpu_path = Path("/proc/cpuinfo")
        meminfo_path = Path("/proc/meminfo")
        version_path = Path("/proc/version")
        osrelease_path = Path("/etc/os-release")
        rocm_path = Path("/opt/rocm/.info/version-utils")
        if cpu_path.exists():
            cpuinfo = cpu_path.read_text()
            cpu_info = search(r"^model name\s*: (.*?)$", cpuinfo)
        else:
            cpuinfo = None
        if meminfo_path.exists():
            meminfo = meminfo_path.read_text()
            ram = search(r"MemTotal:\s*(\S*)", meminfo)
        else:
            meminfo = None
        if version_path.exists():
            version = version_path.read_text()
            kernel_version = search(r"version (\S*)", version)
        else:
            version = None
        if osrelease_path.exists():
            os_release = osrelease_path.read_text()
            distro = search(r'PRETTY_NAME="(.*?)"', os_release)
        else:
            os_release = None
        if rocm_path.exists():
            rocm_info = rocm_path.read_text()
        else:
            rocm_info = None
        try:
            rocm_smi = run(
                [
                    "rocm-smi",
                    "--showvbios",
                    "--showid",
                    "--showproductname",
                    "--showperflevel",
                    "--showclocks",
                    "--showmeminfo",
                    "vram",
                    "lshw",
                ]
            )
        except FileNotFoundError as e:
            rocm_smi = None

        try:
            rocminfo = run(["rocminfo"])
        except FileNotFoundError as e:
            rocminfo = None

        device = rf"^GPU\[{devicenum}\]\s*: "
        hostname = socket.gethostname()

        if rocm_info is None:
            rocm_version = None
        else:
            rocm_version = rocm_info.strip()

        if rocm_smi is None:
            vbios_version = (
                gpuid
            ) = deviceinfo = vram = performance_level = memory_clk = system_clk = None
        else:
            if device is None:
                device = ""
            vbios_version = search(device + r"VBIOS version: (.*?)$", rocm_smi)
            gpuid = search(device + r"Device ID: (.*?)$", rocm_smi)
            if not gpuid:
                gpuid = search(device + r"GPU ID: (.*?)$", rocm_smi)
            vram = search(device + r".... Total Memory .B.: (\d+)$", rocm_smi)
            performance_level = search(device + r"Performance Level: (.*?)$", rocm_smi)
            memory_clk = search(device + r"mclk.*\((.*?)\)$", rocm_smi)
            system_clk = search(device + r"sclk.*\((.*?)\)$", rocm_smi)

        architecture_name, internal_product_name = get_device_info()

        if architecture_name and gpuid:
            deviceinfo = f"{architecture_name.strip()}_{gpuid.strip()}"
        else:
            deviceinfo = None

        if ram is not None:
            ram = "{:.2f} GiB".format(float(ram) / 1024**2)
        if vram is not None:
            vram = "{:.2f} GiB".format(float(vram) / 1024**3)

        gpu_info = get_gpu_info(devicenum)
        cuda_info = get_cuda_info()

        specs = [
            "hostinfo :\n",
            f"hostname : {hostname}",
            f"cpu_info : {cpu_info}",
            f"sbios_info: {get_sbios_info()}",
            f"ram : {ram}",
            f"distro : {distro}",
            f"kernel_version : {kernel_version}",
            f"rocm_version : {rocm_version}",
            "\ndevice info:\n",
            f"device : {deviceinfo}",
            f"vbios_version : {vbios_version}",
            f"performance_level : {performance_level}",
            f"memory_clk : {memory_clk}",
            f"system_clk : {system_clk}",
            f"vram : {vram}",
            f"NVIDIA GPU info: {gpu_info}",
            f"CUDA info: {cuda_info}",
        ]

        if filename == "":
            specs_dict = {
                "hostname": hostname,
                "cpu_info": cpu_info,
                "sbios_info": get_sbios_info(),
                "ram": ram,
                "distro": distro,
                "kernel_version": kernel_version,
                "rocm_version": rocm_version,
                "device": deviceinfo,
                "vbios_version": vbios_version,
                "performance_level": performance_level,
                "memory_clk": memory_clk,
                "system_clk": system_clk,
                "vram": vram,
                "NVIDIA GPU info": gpu_info,
                "CUDA info": cuda_info,
            }
            return specs_dict

        filename = writing_to_file(filename, specs)

    except Exception as error:
        print(error)
        return filename

    return filename


def writing_to_file(filename: str, specs: list) -> str:

    """
    This method is used to store the specs info in the specified location as a text file.

    Args
    ----------
    filename : Absolute filepath where the file is to be created

    Returns
    -------
    filename : Absolute filepath of the created specs file

    Raises
    ------
    IOError (Is a directory error) : If the specified filename ( input ) should be a file
                                     and not a directory


    """

    try:

        # After getting all the values , then those are written in a file.
        with open(filename, "w") as file:
            for info in specs:
                file.write(info + "\n")

    except Exception as error:
        print(error)
        return filename

    return filename



# Copyright (C) 2024-2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Bench launch utils."""

import os
import logging
import pathlib
import asyncio
import sys
from collections import defaultdict
from typing import Dict
from asyncio.subprocess import PIPE, STDOUT

#####################################
# for hipblaslt-bench, can use --yaml
#####################################
def run_bench(benchExecutable,
              probYamlFolder,
              benchType,
              argsDict:Dict[str, str],
              verbose=False,
              timeout=300):
    """Run bench"""
    cmd = [pathlib.Path(benchExecutable).resolve()]

    for argKey, argValue in argsDict.items():
        if len(argValue) != 0:
            if argKey == "--yaml":
                argValue = pathlib.Path(os.path.join(probYamlFolder, argValue)).resolve()
            cmd += [argKey, argValue]
        else:
            cmd += [argKey]

    cmd = [str(x) for x in cmd]
    logging.info('hipblaslt-perf: ' + ' '.join(cmd))
    if verbose:
        print('hipblaslt-perf: ' + ' '.join(cmd))

    startingToken = "["
    csvKeys = ''
    benchResultsList = []
    capturingValues = False
    isAPIOverhead = False

    async def run_command(*args, benchType=benchType, timeout=None):

        process = await asyncio.create_subprocess_exec(
            *args, stdout=asyncio.subprocess.PIPE)

        nonlocal startingToken
        nonlocal csvKeys
        nonlocal benchResultsList
        nonlocal capturingValues
        nonlocal isAPIOverhead

        while True:
            try:
                line = await asyncio.wait_for(process.stdout.readline(),
                                              timeout)
            except asyncio.TimeoutError:
                logging.info(
                    "timeout expired. killed. Please check the process.")
                print("timeout expired. killed. Please check the process.")
                process.kill()  # Timeout or some criterion is not satisfied
                break

            if not line:
                break
            else:
                line = line.decode('utf-8').rstrip('\n')
                line = line.strip()
                if capturingValues:
                    if not line.startswith(benchType): # filter out some irrelative msg
                        # print('irrelative msg:',line)
                        continue
                    print(line)
                    dd_output = defaultdict(str, zip(csvKeys, line.split(',')))
                    benchResultsList += [dd_output]
                    # if is doing api-overhead, the return log will contain serveral values lines
                    capturingValues = True if isAPIOverhead else False
                elif line.startswith(startingToken):
                    line = line.replace('hipblaslt-Gflops', 'gflops')
                    line = line.replace('hipblaslt-GB/s', 'GB/s')
                    splitLine = line.split(':')
                    funcType = splitLine[0]
                    keys = splitLine[1]
                    print(f'\n{keys}')
                    csvKeys = keys.split(',')
                    capturingValues = True
                    isAPIOverhead = (funcType == '[overhead]')
        return await process.wait()  # Wait for the child process to exit

    if sys.platform == "win32":
        loop = asyncio.ProactorEventLoop()  # For subprocess' pipes on Windows
        asyncio.set_event_loop(loop)
    else:
        loop = asyncio.new_event_loop()

    returncode = loop.run_until_complete(run_command(*cmd, timeout=timeout))
    success = returncode == 0

    loop.close()

    return csvKeys, benchResultsList, success


#####################################
# For ./hipblaslt-perf --run_sh
#####################################
def run_sh_cmd(cmdLine,
               verbose=False,
               timeout=300):
    """Run single bench from sh"""

    cmd = cmdLine.split(' ')
    cmd = [str(x) for x in cmd]
    logging.info('running: ' + ' '.join(cmd))
    if verbose:
        print('running: ' + ' '.join(cmd))

    startingToken = "["
    solNameToken = "--Solution name:"
    csvKeys = []
    benchResultsList = {}
    capturingValues = False

    async def run_command(*args, timeout=None):

        process = await asyncio.create_subprocess_exec(
            *args, stdout=asyncio.subprocess.PIPE)

        nonlocal startingToken
        nonlocal csvKeys
        nonlocal benchResultsList
        nonlocal capturingValues

        singleValuesList = []
        solutionName = "N/A"

        while True:
            try:
                line = await asyncio.wait_for(process.stdout.readline(),
                                              timeout)
            except asyncio.TimeoutError:
                logging.info(
                    "timeout expired. killed. Please check the process.")
                print("timeout expired. killed. Please check the process.")
                process.kill()  # Timeout or some criterion is not satisfied
                break

            if not line:
                break
            else:
                line = line.decode('utf-8').rstrip('\n')
                line = line.strip()

                # capturing values right after capturing keys
                if capturingValues:
                    singleValuesList = line.split(',')
                    # default is empty if --print_kernel_info is not in the bench cmd
                    solutionName = "N/A"
                    capturingValues = False
                    continue

                if line.startswith(startingToken):
                    line = line.replace('hipblaslt-Gflops', 'gflops')
                    line = line.replace('hipblaslt-GB/s', 'GB/s')
                    splitLine = line.split(':')
                    # SSN = splitLine[0] # should be [0]
                    keys = splitLine[1] + str(",solution-name")
                    csvKeys = keys.split(',')
                    # print(f'\n{keys}')
                    # print(f'\n{csvKeys}')
                    capturingValues = True # Next line must be values
                else:
                    # if is "--Solution name:" (--print_kernel_info), then we capture this
                    if line.startswith(solNameToken):
                        splitLine = line.split(':')
                        solutionName = splitLine[1].strip()
                    # simply ignore irrelative msg
                    else:
                        continue

        singleValuesList.append(solutionName)
        benchResultsList = defaultdict(str, zip(csvKeys, singleValuesList))

        return await process.wait()  # Wait for the child process to exit

    if sys.platform == "win32":
        loop = asyncio.ProactorEventLoop()  # For subprocess' pipes on Windows
        asyncio.set_event_loop(loop)
    else:
        loop = asyncio.new_event_loop()

    returncode = loop.run_until_complete(run_command(*cmd, timeout=timeout))
    success = returncode == 0

    loop.close()

    return csvKeys, benchResultsList, success



#!/usr/bin/python3
# ########################################################################
# Copyright (C) 2022-2023 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ########################################################################

#Expand hipBLASLt YAML test data file into binary Arguments records"""

import re
import sys
import os
import argparse
import ctypes
from fnmatch import fnmatchcase
from typing import List, Tuple
try:  # Import either the C or pure-Python YAML parser
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader
import yaml

# Regex for type names in the YAML file. Optional *nnn indicates array.
TYPE_RE = re.compile(r'[a-z_A-Z]\w*(:?\s*\*\s*\d+)?$')

# Regex for integer ranges A..B[..C]
INT_RANGE_RE = re.compile(
    r'\s*(-?\d+)\s*\.\.\s*(-?\d+)\s*(?:\.\.\s*(-?\d+)\s*)?$')

# Regex for include: YAML extension
INCLUDE_RE = re.compile(r'include\s*:\s*([-.\w/]+)')

args = {}
testcases = set()
datatypes = {}
param = {}


def main():
    args.update(parse_args().__dict__)
    for doc in get_yaml_docs():
        process_doc(doc)


def process_doc(doc):
    """Process one document in the YAML file"""

    # Ignore empty documents
    if not doc or not doc.get('Tests'):
        return

    # Clear datatypes and params from previous documents
    datatypes.clear()
    param.clear()

    # Return dictionary of all known datatypes
    datatypes.update(get_datatypes(doc))

    # Arguments structure corresponding to C/C++ structure
    param['Arguments'] = type('Arguments', (ctypes.Structure,),
                              {'_fields_': get_arguments(doc)})

    # Special names which get expanded as lists of arguments
    param['dict_lists_to_expand'] = doc.get('Dictionary lists to expand') or ()

    # Lists which are not expanded
    param['lists_to_not_expand'] = doc.get('Lists to not expand') or ()

    # Defaults
    defaults = doc.get('Defaults') or {}

    # Known Bugs
    param['known_bugs'] = doc.get('Known bugs') or []

    # Functions
    param['Functions'] = doc.get('Functions') or {}

    # Instantiate all of the tests, starting with defaults
    for test in doc['Tests']:
        case = defaults.copy()
        case.update(test)
        generate(case, instantiate)


def parse_args():
    """Parse command-line arguments, returning input and output files"""
    parser = argparse.ArgumentParser(description="""
Expand hipBLASLt YAML test data file into binary Arguments records
""")
    parser.add_argument('infile',
                        nargs='?',
                        type=argparse.FileType('r'),
                        default=sys.stdin)
    parser.add_argument('-o', '--out',
                        dest='outfile',
                        type=argparse.FileType('wb'),
                        default=sys.stdout)
    parser.add_argument('-I',
                        help="Add include path",
                        action='append',
                        dest='includes',
                        default=[])
    parser.add_argument('-t', '--template',
                        type=argparse.FileType('r'))
    return parser.parse_args()


def read_yaml_file(file):
    """Read the YAML file, processing include: lines as an extension"""
    file_dir = os.path.dirname(file.name) or os.getcwd()
    source = []
    for line_no, line in enumerate(file, start=1):
        # Keep track of file names and line numbers for each line of YAML
        match = line.startswith('include') and INCLUDE_RE.match(line)
        if not match:
            source.append([line, file.name, line_no])
        else:
            include_file = match.group(1)
            include_dirs = [file_dir] + args['includes']
            for path in include_dirs:
                path = os.path.join(path, include_file)
                if os.path.exists(path):
                    source.extend(read_yaml_file(open(path, 'r')))
                    break
            else:
                sys.exit("In file " + file.name + ", line " +
                         str(line_no) + ", column " + str(match.start(1)+1) +
                         ":\n" + line.rstrip() + "\n" + " " * match.start(1) +
                         "^\nCannot open " + include_file +
                         "\n\nInclude paths:\n" + "\n".join(include_dirs))
    file.close()
    return source


def get_yaml_docs():
    """Parse the YAML file"""
    source = read_yaml_file(args['infile'])

    if args.get('template'):
        source = read_yaml_file(args['template']) + source

    source_str = ''.join([line[0] for line in source])

    def mark_str(mark):
        line = source[mark.line]
        return("In file " + line[1] + ", line " + str(line[2]) + ", column " +
               str(mark.column + 1) + ":\n" + line[0].rstrip() + "\n" +
               ' ' * mark.column + "^\n")

    # We iterate through all of the documents to properly diagnose errors,
    # because the load_all generator does not handle exceptions correctly.
    docs = []
    load = Loader(source_str)
    while load.check_data():
        try:
            doc = load.get_data()
        except yaml.YAMLError as err:
            sys.exit((mark_str(err.problem_mark) if err.problem_mark else "") +
                     (err.problem + "\n" if err.problem else "") +
                     (err.note + "\n" if err.note else ""))
        else:
            docs.append(doc)
    return docs


def get_datatypes(doc):
    """ Get datatypes from YAML doc"""
    dt = ctypes.__dict__.copy()
    for declaration in doc.get('Datatypes') or ():
        for name, decl in declaration.items():
            if isinstance(decl, dict):
                # Create derived class type based on bases and attr entries
                dt[name] = type(name,
                                tuple([eval(t, dt)
                                       for t in decl.get('bases') or ()
                                       if TYPE_RE.match(t)]
                                      ), decl.get('attr') or {})
                # Import class' attributes into the datatype namespace
                for subtype in decl.get('attr') or {}:
                    if TYPE_RE.match(subtype):
                        dt[subtype] = eval(name+'.'+subtype, dt)
            elif isinstance(decl, str) and TYPE_RE.match(decl):
                dt[name] = dt[decl]
            else:
                sys.exit("Unrecognized data type "+name+": "+repr(decl))
    return dt


def get_arguments(doc):
    """The kernel argument list, with argument names and types"""
    return [(var, eval(decl[var], datatypes))
            for decl in doc.get('Arguments') or ()
            if len(decl) == 1
            for var in decl
            if TYPE_RE.match(decl[var])]


def setkey_product(test, key, vals):
    """Helper for setdefaults. Tests that all values in vals is present
    in test, if so then sets test[key] to product of all test[vals]."""
    if all(x in test for x in vals):
        result = 1
        for x in vals:
            if x in ('incx', 'incy'):
                result *= abs(test[x])
            else:
                result *= test[x]
        test[key] = int(result)


def setdefaults(test):
    """Set default values for parameters"""
    # Do not put constant defaults here -- use hipblaslt_common.yaml for that.
    # These are only for dynamic defaults

    if test['transA'] == '*' or test['transB'] == '*':
        test.setdefault('lda', 0)
        test.setdefault('ldb', 0)
        test.setdefault('ldc', 0)
        test.setdefault('ldd', 0)
        test.setdefault('lde', 0)
    elif not test['grouped_gemm']: #catered to gemm default behavior
        M = test['M'][0] if isinstance(test['M'], list) else test['M']
        N = test['N'][0] if isinstance(test['N'], list) else test['N']
        K = test['K'][0] if isinstance(test['K'], list) else test['K']
        test.setdefault('lda', (M if M != 0 else 1) if test['transA'].upper() == 'N'
                        else K if K != 0 else 1)
        test.setdefault('ldb', (K if K != 0 else 1) if test['transB'].upper() == 'N'
                        else N if N != 0 else 1)
        test.setdefault('ldc', M if M != 0 else 1)
        test.setdefault('ldd', M if M != 0 else 1)
        test.setdefault('lde', M if M != 0 else 1)
        if test['batch_count'] > 0:
            test.setdefault('stride_a', test['lda'] *
                            (K if test['transA'].upper() == 'N' else
                             M))
            test.setdefault('stride_b', test['ldb'] *
                            (N if test['transB'].upper() == 'N' else
                             K))
            test.setdefault('stride_c', test['ldc'] * N)
            test.setdefault('stride_d', test['ldd'] * N)
            test.setdefault('stride_e', test['lde'] * N)
            return
    else:
        Ms = test['M']
        Ns = test['N']
        Ks = test['K']
        test.setdefault('lda', [(M if M != 0 else 1) if test['transA'].upper() == 'N'
                        else K if K != 0 else 1 for M, K in zip(Ms, Ks)])
        test.setdefault('ldb', [(K if K != 0 else 1) if test['transB'].upper() == 'N'
                        else N if N != 0 else 1 for K, N in zip(Ks, Ns)])
        test.setdefault('ldc', [M if M != 0 else 1 for M in Ms])
        test.setdefault('ldd', [M if M != 0 else 1 for M in Ms])
        test.setdefault('lde', [M if M != 0 else 1 for M in Ms])

        if test['batch_count'] > 0:
            test.setdefault('stride_a', [lda * K if test['transA'].upper() == 'N' else lda * M for K, M, lda in zip(Ks, Ms, test['lda'])])
            test.setdefault('stride_b',
                            [N * ldb if test['transB'].upper() == 'N' else
                             K * ldb for N, K, ldb in zip(Ns, Ks, test['ldb'])])
            test.setdefault('stride_c', [ldc * N for ldc, N in zip(test['ldc'], Ns)])
            test.setdefault('stride_d', [ldd * N for ldd, N in zip(test['ldd'], Ns)])
            test.setdefault('stride_e', [lde * N for lde, N in zip(test['lde'], Ns)])
            return


    test.setdefault('stride_a', 0)
    test.setdefault('stride_b', 0)
    test.setdefault('stride_c', 0)
    test.setdefault('stride_d', 0)
    test.setdefault('stride_e', 0)


def write_signature(out):
    """Write the signature used to verify binary file compatibility"""
    if 'signature_written' not in args:
        sig = 0
        byt = bytearray("hipBLASLt", 'utf_8')
        byt.append(0)
        last_ofs = 0
        for (name, ctype) in param['Arguments']._fields_:
            member = getattr(param['Arguments'], name)
            for i in range(0, member.offset - last_ofs):
                byt.append(0)
            for i in range(0, member.size):
                byt.append(sig ^ i)
            sig = (sig + 89) % 256
            last_ofs = member.offset + member.size
        for i in range(0, ctypes.sizeof(param['Arguments']) - last_ofs):
            byt.append(0)
        byt.extend(bytes("HIPblaslT", 'utf_8'))
        byt.append(0)
        out.write(byt)
        args['signature_written'] = True

def gen_shape_combinations(M: List, N: List, K: List) -> Tuple[List, List, List]:
    """Generate all shape combinations. This suitable for grouped gemm case."""
    from itertools import product

    if not isinstance(M, List):
        M = [M,]

    if not isinstance(N, List):
        N = [N,]

    if not isinstance(K, List):
        K = [K,]

    problem_sizes = list(product(M, N, K))
    return tuple(tuple(s[i] for s in problem_sizes) for i in range(len(problem_sizes[0])))

def write_test(test):
    """Write the test case out to the binary file if not seen already"""

    # For each argument declared in arguments, we generate a positional
    # argument in the Arguments constructor. For strings, we pass the
    # value of the string directly. For arrays, we unpack their contents
    # into the ctype array constructor and pass the ctype array. For
    # scalars, we coerce the string/numeric value into ctype.
    arg = []
    for name, ctype in param['Arguments']._fields_:
        try:
            if issubclass(ctype, ctypes.Array):
                if issubclass(ctype._type_, ctypes.c_char):
                    arg.append(bytes(test[name], 'utf_8'))
                else:
                    try:
                        arg.append(ctype(*test[name]))
                    except TypeError:
                        arg.append(ctype(test[name]))
            elif issubclass(ctype, ctypes.c_char):
                arg.append(bytes(test[name], 'utf_8'))
            else:
                arg.append(ctype(test[name]))
        except TypeError as err:
            sys.exit("TypeError: " + str(err) + " for " + name +
                     ", which has type " + str(type(test[name])) + "\n")
    byt = bytes(param['Arguments'](*arg))
    if byt not in testcases:
        testcases.add(byt)
        write_signature(args['outfile'])
        args['outfile'].write(byt)


def instantiate(test):
    """Instantiate a given test case"""
    test = test.copy()

    if test['grouped_gemm']:
        if 'matrix_size' in test:
            sizes = test.pop('matrix_size')
            for s in ('M', 'N', 'K'):
                test[s] = tuple(l[s] for l in sizes)
        else:
            M, N, K = gen_shape_combinations(test['M'], test['N'], test['K'])
            test['M'] = M
            test['N'] = N
            test['K'] = K

    gsu_vector = [-1 for _ in range(32)]
    if 'gsu_vector' in test:
      gsu_vector[0] = test['gsu_vector']
    else:
      gsu_vector[0] = 0
    test['gsu_vector'] = gsu_vector

    wgm_vector = [-1 for _ in range(32)]
    if 'wgm_vector' in test:
      wgm_vector[0] = test['wgm_vector']
    else:
      wgm_vector[0] = 0
    test['wgm_vector'] = wgm_vector


    # Any Arguments fields declared as enums (a_type, b_type, etc.)
    enum_args = [decl[0] for decl in param['Arguments']._fields_
                 if decl[1].__module__ == '__main__']
    array_value_args = ["gsu_vector", "wgm_vector"]

    try:
        setdefaults(test)

        # For enum arguments, replace name with value
        for typename in enum_args:
            if test[typename] in datatypes:
                test[typename] = datatypes[test[typename]]

        known_bug_platforms = set()

        # Match known bugs
        if test['category'] not in ('known_bug'):
            for bug in param['known_bugs']:
                for key, value in bug.items():
                    if key == 'known_bug_platforms' or key == 'category':
                        continue
                    if key not in test:
                        break
                    if key == 'function':
                        if not fnmatchcase(test[key], value):
                            break
                    # For keys declared as enums, compare resulting values
                    if key in enum_args:
                        if test[key] != datatypes.get(value, value):
                            break
                    # For array value type, treat as known bug if empty array was given
                    elif key in array_value_args:
                        stripped_val = [i for i in test[key] if i >= 0]

                        # default case in test
                        if stripped_val == [0,]:
                            break

                        if len(value) == 0:
                            continue
                        else:
                            if stripped_val != value:
                                break
                    elif test[key] != value:
                        break
                else:
                    # All values specified in known bug match the test case
                    platforms = bug.get('known_bug_platforms', '')

                    # If at least one known_bug_platforms is specified, add
                    # each platform in platforms to known_bug_platforms set
                    if platforms.strip(' :,\f\n\r\t\v'):
                        known_bug_platforms |= set(re.split('[ :,\f\n\r\t\v]+',
                                                   platforms))
                    else:
                        test['category'] = 'known_bug'
                    break

        # Unless category is already set to known_bug or disabled, set
        # known_bug_platforms to a space-separated list of platforms
        test['known_bug_platforms'] = ' ' . join(known_bug_platforms) if test[
            'category'] not in ('known_bug') else ''

        write_test(test)

    except KeyError as err:
        sys.exit("Undefined value " + str(err) + "\n" + str(test))


def generate(test, function):
    """Generate test combinations by iterating across lists recursively"""
    test = test.copy()

    def should_expand(key):
        grouped_gemm_cond = (test['grouped_gemm'] and key not in ('M', 'N', 'K'))
        normal_cond = type(test[key]) in (tuple, list) and\
                      key not in param['lists_to_not_expand'] and\
                      key != 'matrix_size'

        if test['grouped_gemm']:
            return grouped_gemm_cond and normal_cond
        return normal_cond

    def should_expand_dict_lists(argname):
        normal_cond = argname in test and type(test[argname]) in (tuple, list, dict)
        grouped_gemm_cond = argname != 'matrix_size'

        if test['grouped_gemm']:
            return normal_cond and grouped_gemm_cond
        return normal_cond

    # For specially named lists, they are expanded and merged into the test
    # argument list. When the list name is a dictionary of length 1, its pairs
    # indicate that the argument named by its key takes on values paired with
    # the argument named by its value, which is another dictionary list. We
    # process the value dictionaries' keys in alphabetic order, to ensure
    # deterministic test ordering.
    for argname in param['dict_lists_to_expand']:
        if type(argname) == dict:
            if len(argname) == 1:
                arg, target = list(argname.items())[0]
                if arg in test and type(test[arg]) == dict:
                    pairs = sorted(list(test[arg].items()), key=lambda x: x[0])
                    for test[arg], test[target] in pairs:
                        generate(test, function)
                    return
        elif should_expand_dict_lists(argname):#argname in test and type(test[argname]) in (tuple, list, dict):
            # Pop the list and iterate across it
            ilist = test.pop(argname)

            # For a bare dictionary, wrap it in a list and apply it once
            for item in [ilist] if type(ilist) == dict else ilist:
                try:
                    case = test.copy()
                    case.update(item)  # original test merged with each item
                    generate(case, function)
                except TypeError as err:
                    sys.exit("TypeError: " + str(err) + " for " + argname +
                            ", which has type " + str(type(item)) +
                            "\nA name listed in \"Dictionary lists to "
                            "expand\" must be a defined as a dictionary.\n")
            return

    for key in sorted(list(test)):
        # Integer arguments which are ranges (A..B[..C]) are expanded
        if type(test[key]) == str:
            match = INT_RANGE_RE.match(str(test[key]))
            if match:
                for test[key] in range(int(match.group(1)),
                                       int(match.group(2))+1,
                                       int(match.group(3) or 1)):
                    generate(test, function)
                return

        # For sequence arguments, they are expanded into scalars
        elif should_expand(key):
            for test[key] in test[key]:
                generate(test, function)
            return

    # Replace typed function names with generic functions and types
    if 'hipblaslt_function' in test:
        func = test.pop('hipblaslt_function')
        if func in param['Functions']:
            test.update(param['Functions'][func])
        else:
            test['function'] = func.rpartition('hipblaslt_')[2]
        generate(test, function)
        return

    function(test)


if __name__ == '__main__':
    main()



# Copyright (C) 2025 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import statistics
from dataclasses import dataclass, field
from typing import Dict, List

@dataclass
class MeasurementKey:
    name: str = None
    samples: int = 0
    mean: float = 0
    median: float = 0
    values: List[float] = field(init=False)

    def __init__(self, name):
        self.name = name
        self.values = []

    def getMeanName(self):
        assert(self.name != None)
        return 'mean_' + self.name

    def getMedianName(self):
        assert(self.name != None)
        return 'median_' + self.name

    def addSample(self, value):
        self.samples += 1
        self.values.append(float(value))

    def calcMean(self):
        assert(len(self.values) == self.samples)
        assert(self.name != None)
        self.mean = statistics.mean(self.values)
        return self.mean

    def calcMedian(self):
        assert(len(self.values) == self.samples)
        assert(self.name != None)
        self.median = statistics.median(self.values)
        return self.median

    def getValuesStr(self):
        assert(len(self.values) == self.samples)
        assert(self.name != None)
        return ','.join([str(v) for v in self.values])

@dataclass
class BenchSample:
    probKey: str = None
    measurements: Dict[str, MeasurementKey] = field(default_factory=dict)

    def addSampleOfKey(self, key, value):
        if key not in self.measurements:
            self.measurements[key] = MeasurementKey(name=key)
        self.measurements[key].addSample(value)

    def getMeasurementMeanMedianPair(self, key):
        mean = self.measurements[key].calcMean()
        median = self.measurements[key].calcMedian()
        return str(mean) + ',' + str(median)

    def getMeasurementValues(self, key):
        return self.measurements[key].getValuesStr()

    def finalize(self, num_samples, keysOrder):
        assert(self.probKey != None)
        content = self.probKey + ',' + str(num_samples)

        for key in keysOrder:
            assert(self.measurements[key].samples == num_samples)
            content += ',' + self.getMeasurementMeanMedianPair(key)

        for key in keysOrder:
            content += ',' + self.getMeasurementValues(key)

        return content
